/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * This plugin implements the hashtree implementation for virtual
 * monotonic counters, as described in the following paper:
 *
 * L. F. G. Sarmenta, M. van Dijk, C. W. O’Donnell, J. Rhodes, and
 * S. Devadas, “Virtual monotonic counters and count-limited objects
 * using a TPM without a trusted OS,” in Proceedings of the First ACM
 * Workshop on Scalable Trusted Computing, 2006, pp. 27–42.
 *
 * This implementation was written by Stephen Tate (2012), and there
 * are a few differences from Sarmenta et al.'s 2006 paper.  The most
 * significant is that we allow dynamic growth in the number of
 * counters - the original paper stated that the size (height) of the
 * hashtree is set at the beginning, and while subtrees (or the whole
 * tree) can be RESET, it's not possible to change the size of the
 * tree.  We use two different counter creation commands, CREATE and
 * REUSE, to distinguish between adding a new counter - which may be
 * outside of the current hashtree size - and reusing one that is
 * within the currently defined hashtree.  When CREATE is used and the
 * counter number exceeds the size of the current hash tree, an extra
 * level is added to the tree, with the current tree becoming the left
 * subtree of the new tree.
 *
 * Another enhancement is that we allow the user to retrieve a signed
 * copy of the hashtree's root hash.  This allows us to batch virtual
 * monotonic counter attestations, significantly increasing
 * throughput, as described in this paper:
 *     Need a citation for our paper here...
 */

// Important TODO:
// RESET operation has not been finished, and EXECUTE option needs
// to be done...

#include "tpm/tpm_marshalling.h"
#include "tpm/tpm_commands.h"
#include "tpm/tpm_handles.h"
#include "tpm/tpm_structures.h"
#include "tpm/key_functions.h"
#include "tpm/plugins.h"
#include "crypto/sha1.h"

#include "plugin_hashtree.h"

#include <stdio.h>

/*****************  TPM_COUNTER_BLOB definitions *******************/

#define sizeof_TPM_COUNTER_BLOB(s) (4+20+4+20+4)

static int tpm_marshal_TPM_COUNTER_BLOB(BYTE **ptr, UINT32 *ptr_length,
					TPM_COUNTER_BLOB *blob)
{
  if (tpm_marshal_UINT32(ptr, ptr_length, (blob->counterIDpos))
      || tpm_marshal_TPM_NONCE(ptr, ptr_length, &(blob->counterIDrand))
      || tpm_marshal_UINT32(ptr, ptr_length, (blob->countValue))
      || tpm_marshal_TPM_NONCE(ptr, ptr_length, &(blob->data))
      || tpm_marshal_UINT32(ptr, ptr_length, (blob->authDataBlob)))
    return -1;
  return 0;
}

static int tpm_unmarshal_TPM_COUNTER_BLOB(BYTE **ptr, UINT32 *ptr_length,
					  TPM_COUNTER_BLOB *blob)
{
  if (tpm_unmarshal_UINT32(ptr, ptr_length, &(blob->counterIDpos))
      || tpm_unmarshal_TPM_NONCE(ptr, ptr_length, &(blob->counterIDrand))
      || tpm_unmarshal_UINT32(ptr, ptr_length, &(blob->countValue))
      || tpm_unmarshal_TPM_NONCE(ptr, ptr_length, &(blob->data))
      || tpm_unmarshal_UINT32(ptr, ptr_length, &(blob->authDataBlob)))
    return -1;
  return 0;
}


/*****************  TPM_HASHTREE_EXEC_CERT definitions *******************/

#define sizeof_TPM_HASHTREE_EXEC_CERT(s) (1+20 \
  +sizeof_TPM_COUNTER_BLOB(s.newCounterBlob)\
  +4+s.signatureSize)

static int tpm_marshal_TPM_HASHTREE_EXEC_CERT(BYTE **ptr, UINT32 *ptr_length,
					      TPM_HASHTREE_EXEC_CERT *cert)
{
  if (tpm_marshal_BYTE(ptr, ptr_length, (cert->mode))
      || tpm_marshal_TPM_NONCE(ptr, ptr_length, &(cert->nonce))
      || tpm_marshal_TPM_COUNTER_BLOB(ptr, ptr_length, &(cert->newCounterBlob))
      || tpm_marshal_UINT32(ptr, ptr_length, cert->signatureSize))
    return -1;

  if ((cert->signatureSize > 0)
      && tpm_marshal_BYTE_ARRAY(ptr, ptr_length, cert->signature, cert->signatureSize))
    return -1;

  return 0;
}


/**************************************************************************/

/*
 * The "nullhash" functions compute hashes of empty hash (sub)trees.
 * the _init function computes the hash of an empty leaf, and the _advance
 * function takes a hash of an empty subtree and produces the hash of the
 * next level up in an empty tree - in other words, the hash of two copies
 * of the provided subtree hash concatenated together (as left and right
 * subtrees).
 */

static void hashtree_nullhash_init(TPM_DIGEST *digest) {
  size_t blobSize = sizeof_TPM_COUNTER_BLOB(NULL);
  BYTE blobBuffer[blobSize];
  memset(blobBuffer, 0xff, blobSize);

  tpm_sha1_ctx_t sha1_ctx;
  tpm_sha1_init(&sha1_ctx);
  tpm_sha1_update(&sha1_ctx, blobBuffer, blobSize);
  tpm_sha1_final(&sha1_ctx, digest->digest);
}

static void hashtree_nullhash_advance(TPM_DIGEST *digest) {
  tpm_sha1_ctx_t sha1_ctx;
  tpm_sha1_init(&sha1_ctx);
  tpm_sha1_update(&sha1_ctx, digest->digest, 20);
  tpm_sha1_update(&sha1_ctx, digest->digest, 20);
  tpm_sha1_final(&sha1_ctx, digest->digest);
}

/**************************************************************************/

/* TPM_COUNTER_BLOB copy function */

static void hashtree_counterblob_init_counterblob(TPM_COUNTER_BLOB *src, 
						  TPM_COUNTER_BLOB *dest)
{
  dest->counterIDpos = src->counterIDpos;
  memcpy(dest->counterIDrand.nonce, src->counterIDrand.nonce, 20);
  dest->countValue = src->countValue;
  memcpy(dest->data.nonce, src->data.nonce, 20);
  dest->authDataBlob = src->authDataBlob;
}


/**************************************************************************/

/*
 * Functions for persistent data in the hashtree plugin (init and translate
 * internal/external representations).
 */

static BYTE *hashtree_init_native(void)
{
  TPM_HASHTREE_PERMDATA *ret = tpm_malloc(sizeof(TPM_HASHTREE_PERMDATA));

  if (ret == NULL)
    return NULL;

  ret->height = 0;
  hashtree_nullhash_init(&(ret->rootDigest));

  return (BYTE *)ret;
}

static BYTE *hashtree_data_init_native(BYTE *data, UINT32 size)
{
  TPM_HASHTREE_PERMDATA *ret = tpm_malloc(sizeof(TPM_HASHTREE_PERMDATA));
  if (ret == NULL)
    return NULL;

  if (tpm_unmarshal_UINT32(&data, &size, &ret->height)
       || tpm_unmarshal_TPM_DIGEST(&data, &size, &ret->rootDigest)
       || (size != 0)) {
    tpm_free(ret);
    return NULL;
  }
    
  return (BYTE *)ret;
}

static int hashtree_native_init_data(BYTE *native, BYTE **data, UINT32 *size)
{
  const UINT32 sizeofdata = sizeof_TPM_HASHTREE_PERMDATA(NULL);
  TPM_HASHTREE_PERMDATA *native_hashtree = (TPM_HASHTREE_PERMDATA *)native;

  if (native == NULL)
    return -1;

  if ((*data == NULL) || (*size != sizeofdata)) {
    if (*data != NULL)
      tpm_free(*data);
    *data = tpm_malloc(sizeofdata);
    if (*data == NULL)
      return -1;
    *size = sizeofdata;
  }

  BYTE *ptr = *data;
  UINT32 ptrlen = *size;
  tpm_marshal_UINT32(&ptr, &ptrlen, native_hashtree->height);
  tpm_marshal_TPM_DIGEST(&ptr, &ptrlen, &native_hashtree->rootDigest);
  return 0;
}

/* Function table for the persistent data functions */

static TPM_EXTENSION_FUNCTIONS plugin_hashtree_fns = {
  hashtree_init_native,
  hashtree_data_init_native,
  hashtree_native_init_data
};

/**************************************************************************/

static TPM_HASHTREE_PERMDATA *plugin_hashtree_get_permdata(void)
{
  return
    (TPM_HASHTREE_PERMDATA *)plugin_get_permdata(TPM_SIMEXT_PROJ_HASHTREE,
						 &plugin_hashtree_fns);
}


static void hashtree_hash_counterblob(TPM_COUNTER_BLOB *blob, TPM_DIGEST *digest)
{
  size_t blobSize = sizeof_TPM_COUNTER_BLOB(*blob);
  BYTE blobBuffer[blobSize];
  BYTE *ptr = blobBuffer;
  UINT32 len = blobSize;

  if (tpm_marshal_TPM_COUNTER_BLOB(&ptr, &len, blob))
    return;

  tpm_sha1_ctx_t sha1_ctx;
  tpm_sha1_init(&sha1_ctx);
  tpm_sha1_update(&sha1_ctx, blobBuffer, blobSize);
  tpm_sha1_final(&sha1_ctx, digest->digest);
}


static TPM_RESULT TPM_ExecuteHashTree(TPM_KEY_HANDLE aikHandle,
				      BYTE mode,
				      TPM_COUNTER_BLOB *counterBlob,
				      TPM_NONCE *nonce,
				      UINT32 numSteps,
				      TPM_DIGEST *stepInputs,
				      UINT32 cmdSize,
				      BYTE *command,
				      TPM_AUTH *auth1,
				      TPM_HASHTREE_EXEC_CERT *cert)
{
  TPM_HASHTREE_PERMDATA *permdata = plugin_hashtree_get_permdata();
  TPM_RESULT res = TPM_SUCCESS;

  if (permdata == NULL)
    return TPM_FAIL;

  TPM_KEY_DATA *aik = tpm_get_key(aikHandle);
  if (aik == NULL) return TPM_INVALID_KEYHANDLE;

  /* verify authorization */ 
  if (auth1->authHandle != TPM_INVALID_HANDLE
      || aik->authDataUsage != TPM_AUTH_NEVER) {
    res = tpm_verify_auth(auth1, aik->usageAuth, aikHandle);
    if (res != TPM_SUCCESS) return res;
  }

  // Counter blobs also are supposed to have auth data - need to figure this out

  // numSteps should usually be height-1, except in a CREATE when a new
  // subtree is created, in which case it will be height.  Can check mode
  // now, but can't check the "new subtree" part until after the hash
  // chain verification.

  if ((mode != TPM_HASHTREE_MODE_READROOT) &&
      (numSteps != permdata->height-1) &&
      ((mode != TPM_HASHTREE_MODE_CREATE) || (numSteps != permdata->height)))
    return TPM_BAD_PARAMETER;

  int currPos = counterBlob->counterIDpos;

  TPM_COUNTER_BLOB newCounterBlob;
  hashtree_counterblob_init_counterblob(counterBlob, &newCounterBlob);

  // Initialize cert fields (newcounter initialized below)

  cert->mode = mode;
  memcpy(cert->nonce.nonce, nonce->nonce, 20);
  cert->signatureSize = 0;
  cert->signature = NULL;

  if (mode == TPM_HASHTREE_MODE_READROOT) {
    memcpy(newCounterBlob.data.nonce, permdata->rootDigest.digest, 20);
    hashtree_counterblob_init_counterblob(&newCounterBlob, &cert->newCounterBlob);
  } else {
    if (mode == TPM_HASHTREE_MODE_INCREMENT) {
      debug("Hashtree operation mode: INCREMENT");
      newCounterBlob.countValue++;
      memcpy(newCounterBlob.data.nonce, nonce->nonce, 20);
    } else if ((mode == TPM_HASHTREE_MODE_CREATE) 
	       || (mode == TPM_HASHTREE_MODE_REUSE)) {
      debug("Hashtree operation mode: CREATE/REUSE");
      tpm_get_random_bytes(newCounterBlob.counterIDrand.nonce, 20);
      newCounterBlob.countValue = 0;
      memcpy(newCounterBlob.data.nonce, nonce->nonce, 20);
    }

    hashtree_counterblob_init_counterblob(&newCounterBlob, &cert->newCounterBlob);

    TPM_DIGEST currHash, newHash, nullHash;

    hashtree_nullhash_init(&nullHash);

    if (mode == TPM_HASHTREE_MODE_CREATE)
      memcpy(currHash.digest, nullHash.digest, 20);
    else
      hashtree_hash_counterblob(counterBlob, &currHash);

    hashtree_hash_counterblob(&newCounterBlob, &newHash);

    int newSubTree = 1;

    UINT32 i;
    for (i=0; i<numSteps; i++) {
      if (newSubTree) {
	if (i+1<numSteps) {
	  if (memcmp(nullHash.digest, stepInputs[i].digest, 20) || (currPos&1))
	    newSubTree = 0;
	} else {
	  if (memcmp(permdata->rootDigest.digest, stepInputs[i].digest, 20) 
	      || ((currPos&1) == 0))
	    newSubTree = 0;
	}
      }

      tpm_sha1_ctx_t sha1_ctx_cur;
      tpm_sha1_ctx_t sha1_ctx_new;
      tpm_sha1_init(&sha1_ctx_cur);
      tpm_sha1_init(&sha1_ctx_new);
      if (currPos&1) {
	tpm_sha1_update(&sha1_ctx_cur, stepInputs[i].digest, 20);
	tpm_sha1_update(&sha1_ctx_cur, currHash.digest, 20);
	tpm_sha1_update(&sha1_ctx_new, stepInputs[i].digest, 20);
	tpm_sha1_update(&sha1_ctx_new, newHash.digest, 20);
      } else {
	tpm_sha1_update(&sha1_ctx_cur, currHash.digest, 20);
	tpm_sha1_update(&sha1_ctx_cur, stepInputs[i].digest, 20);
	tpm_sha1_update(&sha1_ctx_new, newHash.digest, 20);
	tpm_sha1_update(&sha1_ctx_new, stepInputs[i].digest, 20);
      }
      tpm_sha1_final(&sha1_ctx_cur, currHash.digest);
      tpm_sha1_final(&sha1_ctx_new, newHash.digest);
      currPos >>= 1;
      hashtree_nullhash_advance(&nullHash);
    }

    if (!newSubTree && memcmp(currHash.digest, permdata->rootDigest.digest, 20))
      return TPM_BAD_COUNTER;

    if (newSubTree && (mode == TPM_HASHTREE_MODE_CREATE))
      permdata->height++;

    if ((mode == TPM_HASHTREE_MODE_INCREMENT)
	|| (mode == TPM_HASHTREE_MODE_CREATE)
	|| (mode == TPM_HASHTREE_MODE_RESET)) {
      memcpy(permdata->rootDigest.digest, newHash.digest, 20);
    }
  }

  // If we got to this point, all is OK, so sign the exec certificate

  size_t areaToSignSize = 1+20+sizeof_TPM_COUNTER_BLOB(cert->newCounterBlob);
  BYTE areaToSign[areaToSignSize];

  UINT32 len = areaToSignSize;
  BYTE *pos = areaToSign;

  if (tpm_marshal_BYTE(&pos, &len, cert->mode)
      || tpm_marshal_TPM_NONCE(&pos, &len, &cert->nonce)
      || tpm_marshal_TPM_COUNTER_BLOB(&pos, &len, &cert->newCounterBlob))
    return TPM_FAIL;   // This really should never fail...

  /* sign the exec certificate */

  return key_keydata_sign(aik, areaToSign, areaToSignSize, TRUE,
			  &cert->signature, &cert->signatureSize);
}


/*
 * This function unmarsals the request, calls the real work function,
 * and marshals the results back up.
 */
static TPM_RESULT execute_TPM_ExecuteHashTree(TPM_REQUEST *req, TPM_RESPONSE *rsp)
{
  BYTE *ptr;
  UINT32 len;
  TPM_KEY_HANDLE aikHandle;
  BYTE mode;
  TPM_COUNTER_BLOB counterBlob;
  TPM_NONCE nonce;
  UINT32 i;
  UINT32 numSteps;
  TPM_DIGEST *stepInputs = NULL;
  UINT32 cmdSize;
  BYTE *command = NULL;
  TPM_HASHTREE_EXEC_CERT cert;
  TPM_RESULT res = TPM_SUCCESS;
  info("In execute_TPM_ExecuteHashTree");
  tpm_compute_in_param_digest(req);
  ptr = req->param;
  len = req->paramSize;

  if (tpm_unmarshal_TPM_KEY_HANDLE(&ptr, &len, &aikHandle)
      || tpm_unmarshal_BYTE(&ptr, &len, &mode)
      || tpm_unmarshal_TPM_COUNTER_BLOB(&ptr, &len, &counterBlob)
      || tpm_unmarshal_TPM_NONCE(&ptr, &len, &nonce)
      || tpm_unmarshal_UINT32(&ptr, &len, &numSteps))
    return TPM_BAD_PARAMETER;

  if (numSteps > 32)
      return TPM_BAD_PARAMETER;

  stepInputs = tpm_malloc(numSteps*sizeof(TPM_DIGEST));
  if (stepInputs == NULL)
    return TPM_FAIL;

  for (i=0; i<numSteps; i++) {
      if (tpm_unmarshal_TPM_DIGEST(&ptr, &len, &stepInputs[i])) {
	  tpm_free(stepInputs);
	  return TPM_BAD_PARAMETER;
      }
  }

  if (tpm_unmarshal_UINT32(&ptr, &len, &cmdSize)) {
    tpm_free(stepInputs);
    return TPM_BAD_PARAMETER;
  }
  if (len != 0) {
    tpm_free(stepInputs);
    return TPM_BAD_PARAMETER;
  }
  /* execute command */
  res = TPM_ExecuteHashTree(aikHandle, mode, &counterBlob, &nonce, numSteps,
			    stepInputs, cmdSize, command, &req->auth1, &cert);

  tpm_free(stepInputs);
  if (res != TPM_SUCCESS) return res;

  rsp->paramSize = len = sizeof_TPM_HASHTREE_EXEC_CERT(cert);
  rsp->param = ptr = tpm_malloc(len);
  if (ptr == NULL
      || tpm_marshal_TPM_HASHTREE_EXEC_CERT(&ptr, &len, &cert)) {
    tpm_free(rsp->param);
    res = TPM_FAIL;
  }

  tpm_free(cert.signature);
  return res;
}


/*
 * The TPM_ExecuteHashTree command is the first ordinal in the 
 * PROJ_HASHTREE project
 */

#define TPM_ORD_ExecuteHashTree TPM_SIMEXT_ORD_PROJ(TPM_SIMEXT_PROJ_HASHTREE,1)


/*
 * Code to register ...
 */

static commands_info tpm_hashtree_cmds[] = {
  {TPM_ORD_ExecuteHashTree,"TPM_ORD_ExecuteHashTree",execute_TPM_ExecuteHashTree,4,0,0}
};

 
void plugin_init(void)
{
  commands_add_set(sizeof(tpm_hashtree_cmds)/sizeof(commands_info),
		   tpm_hashtree_cmds);
}

