/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 *  HMAC keytype functions
 *
 *  This code is to support extending TPMs with the capability of
 *  performing HMAC with a secret TPM key.  While there could be other
 *  uses for this, the main purpose of this plugin is to experiment
 *  with the ideas in this paper:
 *
 *      V. Gunupudi and S. R. Tate. "Random Oracle Instantiation in
 *      Distributed Protocols Using Trusted Platform Modules," 3rd
 *      IEEE International Symposium on Security in Networks and
 *      Distributed Systems (SSNDS), proceedings as part of the 21st
 *      International Conference on Advanced Information Networking
 *      and Applications Workshops, 2007, pp. 463-469.
 *
 *  This paper is currently available at
 *  http://www.uncg.edu/cmp/faculty/srtate/publications.php
 *
 *  HMAC/oracles code originally written in 2012 by Sahana Rangaraju,
 *  with subsequent modifications by Stephen Tate.
 */

#include "crypto/hmac.h"
#include "crypto/rc4.h"
#include "tpm/tpm_marshalling.h"
#include "tpm/tpm_commands.h"
#include "tpm/tpm_structures.h"
#include "tpm/tpm_handles.h"
#include "tpm/plugins.h"

#include "plugin_randoracle.h"

/* A macro to make some things simpler... */

#define BITSTOBYTES(bits) ((bits+7)/8)


/* An HMAC key is special-purpose, and can't be used for Sign, Bind, etc. */

static BOOL hmac_is_genkey(void)
{
  return FALSE;
}


/* An HMAC key is special-purpose, and can't be used for transport fns */

static BOOL hmac_is_transkey(void)
{
  return FALSE;
}


/* An HMAC key is special-purpose, and can't be used for transport fns */

static BOOL hmac_es_is_validtrans(TPM_ENC_SCHEME es)
{
  return FALSE;
}


static BOOL hmac_parms_is_supported(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return hmac->keyLength == TPM_HMAC_KEY_LENGTH;
}


static BOOL hmac_parms_is_minimal_size(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return hmac->keyLength == TPM_HMAC_KEY_LENGTH;
}


static BOOL hmac_parms_is_fips_size(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return hmac->keyLength == TPM_HMAC_KEY_LENGTH;
}


static BOOL hmac_parms_is_strong_size(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return hmac->keyLength == TPM_HMAC_KEY_LENGTH;
}


/* HMAC keys are not used for encryption... */

static BOOL hmac_parms_is_good_es(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


static BOOL hmac_parms_is_any_es(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


/* HMAC keys are not used for signing... */

static BOOL hmac_parms_is_good_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


static BOOL hmac_parms_is_idkey_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


static BOOL hmac_parms_is_any_ss(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


static BOOL hmac_parms_requires_info(TPM_KEY_PARMS *parms)
{
  return FALSE;
}


static BOOL hmac_key_is_valid_keylengths(TPM_KEY *key, TPM_STORE_ASYMKEY *store)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)key->algorithmParms.bparms;
  return hmac->keyLength == TPM_HMAC_KEY_LENGTH;
}


static UINT32 hmac_parms_get_ctextsize(TPM_KEY_PARMS *parms)
{
  return 0;
}


static UINT32 hmac_key_parms_get_sigsize(TPM_KEY_PARMS *parms)
{
  return SHA1_DIGEST_LENGTH;
}


static UINT32 hmac_parms_get_ptextsizemax(TPM_KEY_PARMS *parms)
{
  return 0;
}


static void hmac_parms_set_strong_keylen(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  hmac->keyLength = TPM_HMAC_KEY_LENGTH;
}


static void hmac_parms_set_strong_es(TPM_KEY_PARMS *parms)
{
  parms->encScheme = TPM_ES_NONE;
  parms->sigScheme = TPM_SS_NONE;
}


static int hmac_parms_set_default_algparms(TPM_KEY_PARMS *parms)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  parms->parmSize = sizeof(TPM_HMAC_KEY_PARMS);
  hmac->keyLength = TPM_HMAC_KEY_LENGTH;
  return 0;
}


static int hmac_parms_init_algparms(TPM_KEY_PARMS *src, BYTE *destalgparms)
{
  TPM_HMAC_KEY_PARMS *srchmac = (TPM_HMAC_KEY_PARMS *)src->bparms;
  TPM_HMAC_KEY_PARMS *desthmac = (TPM_HMAC_KEY_PARMS *)destalgparms;
  desthmac->keyLength = srchmac->keyLength;
  return 0;
}


static int hmac_parms_native_init_storepubkey(TPM_KEY_PARMS *parms, 
					 BYTE *native, TPM_STORE_PUBKEY *out)
{
  out->keyLength = 0;
  out->key = NULL;
  return 0;
}
    

static int hmac_parms_nativepk_init_storepubkey(TPM_KEY_PARMS *parms, 
					 BYTE *nativepk, TPM_STORE_PUBKEY *out)
{
  out->keyLength = 0;
  out->key = NULL;
  return 0;
}


static int hmac_parms_native_init_storeprivkey(TPM_KEY_PARMS *parms,
				   BYTE *nativeKey, TPM_STORE_PRIVKEY *privKey)
{
  tpm_hmac_key_t *hmackey = (tpm_hmac_key_t *)nativeKey;
  TPM_HMAC_KEY_PARMS *hmacparms = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  privKey->keyLength = BITSTOBYTES(hmacparms->keyLength);
  privKey->key = tpm_malloc(privKey->keyLength);
  if (privKey->key == NULL) {
    privKey->keyLength = 0;
    return -1;			
  }
  memcpy(privKey->key, hmackey->key, privKey->keyLength);
  return 0;
}


static int hmac_parms_storekeys_init_native(TPM_KEY_PARMS *parms, 
				       TPM_STORE_PUBKEY *spubkey,
				       TPM_STORE_PRIVKEY *sprivkey,
				       BYTE *native)
{
  TPM_HMAC_KEY_PARMS *hmacparms = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  tpm_hmac_key_t *hmac = (tpm_hmac_key_t *)native; 
  memcpy(hmac->key, sprivkey->key, BITSTOBYTES(hmacparms->keyLength));
  return 0;
}


static int hmac_parms_native_init_native(TPM_KEY_PARMS *parms, 
					 BYTE *nativein,
					 BYTE *nativeout)
{
  TPM_HMAC_KEY_PARMS *hmacparms= (TPM_HMAC_KEY_PARMS *)parms->bparms;
  tpm_hmac_key_t *hmacin = (tpm_hmac_key_t *)nativein;
  tpm_hmac_key_t *hmacout = (tpm_hmac_key_t *)nativeout;
  memcpy(hmacout->key, hmacin->key, BITSTOBYTES(hmacparms->keyLength));
  return 0;
}


static int hmac_pubkey_init_nativepk(TPM_PUBKEY *src, BYTE *nativepk)
{
  return 0;
}


static int hmac_keydata_init_nativepk(TPM_KEY_DATA *key, BYTE *nativepk)
{
  return 0;
}


static int hmac_parms_generate_native(TPM_KEY_PARMS *parms, BYTE *nativeout)
{
 TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
 tpm_hmac_key_t *native = (tpm_hmac_key_t *)nativeout;
 UINT32 keyBytes = BITSTOBYTES(hmac->keyLength);
 tpm_get_random_bytes(native->key, keyBytes);
 return 0;
}


static int hmac_pubkeydata_encrypt(TPM_PUBKEY_DATA *key, BYTE *in,
				   UINT32 in_size, BYTE *enc, UINT32 *enc_size)
{
 return -1;
}


static int hmac_keydata_decrypt(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
                BYTE *out, UINT32 *out_size)
{
  return -1;
}


static TPM_RESULT hmac_parms_native_sign(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
	    const uint8_t *data, size_t data_len, BOOL hashFirst, uint8_t *sig)
{
  return TPM_FAIL;
}


static TPM_RESULT hmac_parms_nativepk_verify(TPM_KEY_PARMS *parms, 
        BYTE *nativePK, BYTE *data, UINT32 data_len, BOOL hashFirst, BYTE *sig)
{
  return TPM_FAIL;
}


static UINT32 hmac_parms_native_marshal_size(TPM_KEY_PARMS *parms, BYTE *native)
{
  TPM_HMAC_KEY_PARMS *hmackey = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return 2 + BITSTOBYTES(hmackey->keyLength);
}


static int hmac_parms_marshal_native(TPM_KEY_PARMS *parms, BYTE **ptr,
			      UINT32 *length, BYTE *native)
{
  TPM_HMAC_KEY_PARMS *hmacparms= (TPM_HMAC_KEY_PARMS *)parms->bparms;
  UINT32 hmacBytes = BITSTOBYTES(hmacparms->keyLength);
  tpm_hmac_key_t *v = (tpm_hmac_key_t *)native;
  if (*length < hmac_parms_native_marshal_size(parms, native)) return -1;
  tpm_marshal_UINT16(ptr, length, hmacparms->keyLength);
  memcpy(*ptr, v->key, hmacBytes);
  *ptr = *ptr + hmacBytes;
  *length= *length - hmacBytes;
  return 0;
}


static int hmac_parms_unmarshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, 
				UINT32 *length, BYTE *native)
{
  tpm_hmac_key_t *hmackey = (tpm_hmac_key_t *)native;
  UINT16 k_len;
  if (tpm_unmarshal_UINT16(ptr, length, &k_len)) return -1;
  if (k_len != BITSTOBYTES(TPM_HMAC_KEY_LENGTH)) return -1;
  if (*length < (UINT32)k_len)
    return -1;
  else
    memcpy(hmackey->key, *ptr, k_len);
  *ptr += k_len;
  *length -= k_len;
  return 0;
}


static UINT32 hmac_parms_nativepk_marshal_size(TPM_KEY_PARMS *parms,
					       BYTE *nativepk)
{
  return 0;
}


static int hmac_parms_marshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr,
			      UINT32 *length, BYTE *nativepk)
{
  return 0;
}


static int hmac_parms_unmarshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
				UINT32 *length, BYTE *nativepk)
{
  return 0;
}


static int hmac_parms_marshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				   UINT32 *length)
{
  TPM_HMAC_KEY_PARMS *hmac = (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return tpm_marshal_UINT32(ptr, length, hmac->keyLength);
}


static int hmac_parms_unmarshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				     UINT32 *length)
{
  TPM_HMAC_KEY_PARMS *hmac= (TPM_HMAC_KEY_PARMS *)parms->bparms;
  return tpm_unmarshal_UINT32(ptr, length, &hmac->keyLength);
}


static void hmac_parms_free(TPM_KEY_PARMS *parms)
{
}


static void hmac_parms_native_free(TPM_KEY_PARMS *parms, BYTE *native)
{
}


static void hmac_parms_nativepk_free(TPM_KEY_PARMS *parms, BYTE *pnativepk)
{
}


static int hmac_selftest(void)
{
  return 0;  // Might put some self-test here later...
}


static int hmac_init(void)
{
  int ret = 0;
  if (sizeof(TPM_HMAC_KEY_PARMS) > TPM_ALG_PARMS_MAXSIZE) {
    info("Can't load HMAC keytype: need %d bytes for KEY_PARMS and only have %d",
	 sizeof(TPM_HMAC_KEY_PARMS), TPM_ALG_PARMS_MAXSIZE);
    ret = -1;
  }
  if (sizeof(tpm_hmac_key_t) > TPM_NATIVE_KEY_MAXSIZE) {
    info("Can't load HMAC keytype: need %d bytes for native and only have %d",
	 sizeof(tpm_hmac_key_t), TPM_NATIVE_KEY_MAXSIZE);
    ret = -1;
  }
  if (ret) {
    info("To make this work, you must recompile the TPM emulator with a bigger buffer");
  }
  return ret;
}


keytype_info hmac_info = {
  TPM_ALG_HMAC,   /* algorithmID */
  "HMAC",
  hmac_init,

  hmac_is_genkey,
  hmac_is_transkey,
  hmac_es_is_validtrans,

  hmac_parms_is_supported,
  hmac_parms_is_minimal_size,
  hmac_parms_is_fips_size,
  hmac_parms_is_strong_size,
  hmac_parms_is_good_es,
  hmac_parms_is_any_es,
  hmac_parms_is_good_ss,
  hmac_parms_is_idkey_ss,
  hmac_parms_is_any_ss,
  hmac_parms_requires_info,
  hmac_key_is_valid_keylengths,

  hmac_parms_get_ctextsize,
  hmac_key_parms_get_sigsize,
  hmac_parms_get_ptextsizemax,

  hmac_parms_set_strong_keylen,
  hmac_parms_set_strong_es,

  hmac_parms_set_default_algparms,
  hmac_parms_init_algparms,

  hmac_parms_native_init_storepubkey,
  hmac_parms_nativepk_init_storepubkey,

  hmac_parms_native_init_storeprivkey,

  hmac_parms_storekeys_init_native,
  hmac_parms_native_init_native,

  hmac_pubkey_init_nativepk,
  hmac_keydata_init_nativepk,

  hmac_parms_generate_native,
  hmac_pubkeydata_encrypt,
  hmac_keydata_decrypt,
  hmac_parms_native_sign,
  hmac_parms_nativepk_verify,

  hmac_parms_native_marshal_size,
  hmac_parms_marshal_native,
  hmac_parms_unmarshal_native,

  hmac_parms_nativepk_marshal_size,
  hmac_parms_marshal_nativepk,
  hmac_parms_unmarshal_nativepk,

  hmac_parms_marshal_algkeyparms,
  hmac_parms_unmarshal_algkeyparms,

  hmac_parms_free,
  hmac_parms_native_free,  
  hmac_parms_nativepk_free,

  hmac_selftest
};

