/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 *  HMAC TPM command extensions code
 *
 *  This code is to support extending TPMs with the capability of
 *  performing HMAC with a secret TPM key.  While there could be other
 *  uses for this, the main purpose of this plugin is to experiment
 *  with the ideas in this paper:
 *
 *      V. Gunupudi and S. R. Tate. "Random Oracle Instantiation in
 *      Distributed Protocols Using Trusted Platform Modules," 3rd
 *      IEEE International Symposium on Security in Networks and
 *      Distributed Systems (SSNDS), proceedings as part of the 21st
 *      International Conference on Advanced Information Networking
 *      and Applications Workshops, 2007, pp. 463-469.
 *
 *  This paper is currently available at
 *  http://www.uncg.edu/cmp/faculty/srtate/publications.php
 *
 *  HMAC/oracles code originally written in 2012 by Sahana Rangaraju,
 *  with subsequent modifications by Stephen Tate.
 */

#include "crypto/hmac.h"
#include "tpm/tpm_marshalling.h"
#include "tpm/tpm_commands.h"
#include "tpm/tpm_handles.h"
#include "tpm/tpm_structures.h"
#include "tpm/key_functions.h"
#include "tpm/plugins.h"

#include "plugin_randoracle.h"


/*
 * This function does the real work of the TPM_HMAC command (note that
 * the name had to have "_Cmd" appended since "TPM_HMAC" is defined
 * elsewhere).  All this function does is use the loaded HMAC key to
 * compute the HMAC of the designated data.
 */
static TPM_RESULT TPM_HMAC_Cmd(TPM_KEY_HANDLE keyHandle, UINT32 areaToMACSize, 
                    BYTE *areaToMAC, TPM_AUTH *auth1,  
                    UINT32 *macSize, BYTE **mac)
{
  TPM_RESULT res;
  TPM_KEY_DATA *key;
  tpm_hmac_ctx_t hmac_ctx;
  info("TPM_HMAC()");
  /* get key */
  key = tpm_get_key(keyHandle);
  if (key == NULL) return TPM_INVALID_KEYHANDLE;
  /* verify authorization */ 
  if (auth1->authHandle != TPM_INVALID_HANDLE
      || key->authDataUsage != TPM_AUTH_NEVER) {
    res = tpm_verify_auth(auth1, key->usageAuth, keyHandle);
    if (res != TPM_SUCCESS) return res;
  }
  if (areaToMACSize == 0) return TPM_BAD_PARAMETER;
  if (key->algorithmParms.algorithmID != TPM_ALG_HMAC)
    return TPM_INVALID_KEYUSAGE;

  tpm_hmac_key_t *hmackey = (tpm_hmac_key_t *)key->bnativekey;
  TPM_HMAC_KEY_PARMS *hmacParms =
    (TPM_HMAC_KEY_PARMS *)key->algorithmParms.bparms;

  *macSize = key_parms_get_sigsize(&key->algorithmParms);
  *mac = tpm_malloc(*macSize);

  tpm_hmac_init(&hmac_ctx, hmackey->key, (hmacParms->keyLength+7)/8);
  tpm_hmac_update(&hmac_ctx, areaToMAC, areaToMACSize);
  tpm_hmac_final(&hmac_ctx, *mac);

  return TPM_SUCCESS;
}


/*
 * This function unmarsals the request, calls the real work function,
 * and marshals the results back up.
 */
static TPM_RESULT execute_TPM_HMAC(TPM_REQUEST *req, TPM_RESPONSE *rsp)
{
  BYTE *ptr;
  UINT32 len;
  TPM_KEY_HANDLE keyHandle;
  UINT32 areaToMACSize;
  BYTE *areaToMAC;
  UINT32 macSize;
  BYTE *mac = NULL;
  TPM_RESULT res = TPM_SUCCESS;
  tpm_compute_in_param_digest(req);
  ptr = req->param;
  len = req->paramSize;
  if (tpm_unmarshal_TPM_KEY_HANDLE(&ptr, &len, &keyHandle)
      || tpm_unmarshal_UINT32(&ptr, &len, &areaToMACSize)
      || tpm_unmarshal_BLOB(&ptr, &len, &areaToMAC, areaToMACSize)
      || len !=0) return TPM_BAD_PARAMETER;
  /* execute command */
  res = TPM_HMAC_Cmd(keyHandle, areaToMACSize, areaToMAC, &req->auth1, &macSize, &mac);
  if (res != TPM_SUCCESS) return res;
  rsp->paramSize = len = 4 + macSize;
  rsp->param = ptr = tpm_malloc(len);
  if (ptr == NULL
      || tpm_marshal_UINT32(&ptr, &len, macSize)
      || tpm_marshal_BLOB(&ptr, &len, mac, macSize)) {
    tpm_free(rsp->param);
    res = TPM_FAIL;
  }
  tpm_free(mac);
  return res;
}


/*
 * Code to register the new TPM_HMAC command and HMAC key type with
 * the simulator.
 */

commands_info tpm_hmac_cmds[] = {
  {TPM_ORD_HMAC,"TPM_ORD_HMAC",execute_TPM_HMAC,4,0,0}
};

 
void plugin_init(void)
{
  keytype_add(&hmac_info);
  commands_add_set(sizeof(tpm_hmac_cmds)/sizeof(commands_info), tpm_hmac_cmds);
}

