/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * Original plugin/extension code by Alex Kilgore and Stephen Tate.
 */

#include <stdio.h>
#include <dlfcn.h>
#include "tpm_structures.h"
#include "key_functions.h"
#include "plugins.h"


#define MAX_KEYTYPES 100

keytype_info *key_typeInfo[MAX_KEYTYPES];
int key_numKeyTypes;

/*
 * Function to register a new keytype
 */
void keytype_add(keytype_info *kt_info)
{
  if ((kt_info != NULL) && (key_numKeyTypes < MAX_KEYTYPES)) {
    if (kt_info->init()) {
      info("Failed to load keytype %s", kt_info->debugname);
    } else {
      key_typeInfo[key_numKeyTypes++] = kt_info;
      info("Successfully loaded keytype %s", kt_info->debugname);
    }
  }
}



/************** Information function ****************/


/**********************************************************************
 * .._is_genkey() means the specified algorithm can be used for Sign, Seal, 
 * Unbind, etc. operations
 */
BOOL key_algorithmID_is_genkey(TPM_ALGORITHM_ID id)
{
  int i;
  for (i=0; i<key_numKeyTypes; i++) {
    if (id == key_typeInfo[i]->algorithmID)
      return key_typeInfo[i]->is_genkey();
  }
  return FALSE;
}


/**********************************************************************
 * .._is_transkey() means the specified algorithm can be used for
 * EstablishTransport, ExecuteTransport, and ReleaseTransportSigned
 */
BOOL key_algorithmID_is_transkey(TPM_ALGORITHM_ID id)
{
  int i;
  for (i=0; i<key_numKeyTypes; i++) {
    if (id == key_typeInfo[i]->algorithmID)
      return key_typeInfo[i]->is_transkey();
  }
  return FALSE;
}

/****
 * NOTE: This is a strange function - just an encryption scheme is sent in,
 * without an accompanying algorithmID.  My interpretation: return true if
 * this is valid ES for ANY supported algorithm.
 */
BOOL key_es_is_validtrans(TPM_ENC_SCHEME es)
{
  int i;
  for (i=0; i<key_numKeyTypes; i++) {
    if (key_typeInfo[i]->es_is_validtrans(es))
      return TRUE;
  }
  return FALSE;
}


/**********************************************************************
 * .._parms_is_supported returns true if the parameters in the parms
 * structure supported?  (key size, etc.)
 */
BOOL key_parms_is_supported(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_supported(parms);
}


/**********************************************************************
 * .._parms_is_minimal_size returns true if the keysize is at least some
 * minimal reasonable value
 */
BOOL key_parms_is_minimal_size(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_minimal_size(parms);
}


/**********************************************************************
 * .._parms_is_fips_size returns true if the keysize is at least some
 * the size required by FIPS requirements (maybe be more than the minimal
 * size otherwise supported).
 */
BOOL key_parms_is_fips_size(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_fips_size(parms);
}


/**********************************************************************
 * .._parms_is_strong_size returns true if the keysize is sufficient to
 * support strong security (e.g., 2048 bits for an RSA key)
 */
BOOL key_parms_is_strong_size(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_strong_size(parms);
}


/**********************************************************************
 * .._parms_is_good_es returns true is the ES is a strong value (can
 * deprecate use of weaker ESes by only recognizing stronger schemes
 * as "good"
 */
BOOL key_parms_is_good_es(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_good_es(parms);
}


/**********************************************************************
 * .._parms_is_any_es returns true if ES is set to any valid encryption
 * scheme for this keytype/algorithm
 */
BOOL key_parms_is_any_es(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_any_es(parms);
}


/**********************************************************************
 * .._parms_is_good_ss returns true is the SS is a strong value (can
 * deprecate use of weaker SSes by only recognizing stronger schemes
 * as "good"
 */
BOOL key_parms_is_good_ss(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_good_ss(parms);
}


/**********************************************************************
 * .._parms_is_idkey_ss returns true iff this SS is a valid signature
 * scheme for an AIK to use
 *
 * Note - needs a separate idkey check - this is something of a
 * strange thing from the TPM_MakeIdentity function.  It doesn't check
 * for this specific signature scheme up front, but the last step
 * requires a signature to be made with this SS.  So we'll just check
 * up front...
 */
BOOL key_parms_is_idkey_ss(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_idkey_ss(parms);
}


/**********************************************************************
 * .._parms_is_any_ss returns true if SS is set to any valid signature
 * scheme for this keytype/algorithm
 */
BOOL key_parms_is_any_ss(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_is_any_ss(parms);
}


/**********************************************************************
 * .._parms_requires_info returns true iff this signature scheme requires
 * the use of a TPM_SIGN_INFO structure
 */
BOOL key_parms_requires_info(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_requires_info(parms);
}


/**********************************************************************
 * .._key_is_valid_keylengths checks public and private key lengths and
 * returns true if they are compatible/agree with each other
 */
BOOL key_key_is_valid_keylengths(TPM_KEY *key, TPM_STORE_ASYMKEY *store)
{
  return key->algorithmParms.parmFns->key_is_valid_keylengths(key, store);
}


/**********************************************************************
 * .._parms_get_ctextsize returns the size of the ciphertext (in bytes)
 * produced by encryption with a key described in parms
 */
UINT32 key_parms_get_ctextsize(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_get_ctextsize(parms);
}


/**********************************************************************
 * .._parms_get_sigsize returns the size of the signature (in bytes)
 * produced by signing with a key described in parms
 */
UINT32 key_parms_get_sigsize(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_get_sigsize(parms);
}


/**********************************************************************
 * .._parms_get_ptextsizemax returns the maximum size of a plaintext
 * (in bytes) supported by an encryption algorithm/scheme described
 * in parms
 */
UINT32 key_parms_get_ptextsizemax(TPM_KEY_PARMS *parms)
{
  return parms->parmFns->parms_get_ptextsizemax(parms);
}


/*********  Basic parameter setting funcitons  ***********/

/**********************************************************************
 * .._parms_set_strong_keylen set the keylength in parms to be a size
 * that supports strong security (e.g., 2048 bits for an RSA key)
 */
void key_parms_set_strong_keylen(TPM_KEY_PARMS *parms)
{
  parms->parmFns->parms_set_strong_keylen(parms);
}


/**********************************************************************
 * .._parms_set_strong_es sets the parms encryption scheme to be a scheme
 * that provides strong security
 */
void key_parms_set_strong_es(TPM_KEY_PARMS *parms)
{
  parms->parmFns->parms_set_strong_es(parms);
}

/*********  Structure initialization functions  **********/


/**********************************************************************
 * .._algorithmID_init_parms initializes the parms structure using
 * default values for the specified keytype/algorithm
 */
int key_algorithmID_init_parms(TPM_ALGORITHM_ID id, TPM_KEY_PARMS *parms)
{
  int i;
  parms->encScheme = TPM_ES_NONE;
  parms->sigScheme = TPM_SS_NONE;
  parms->parmSize = 0;
  for (i=0; i<key_numKeyTypes; i++) {
    if (id == key_typeInfo[i]->algorithmID) {
      parms->algorithmID = id;
      parms->parmFns = key_typeInfo[i];
      if (!parms->parmFns->parms_set_default_algparms(parms))
	return 0;
      break;
    }
  }
  parms->algorithmID = 0;
  parms->parmFns = key_typeInfo[0];  /* Null functions */
  parms->parmSize = 0;
  return -1;
}


/**********************************************************************
 * .._parms_init_parms initializes the parms structure by copying info
 * from src parms structure
 */
int key_parms_init_parms(TPM_KEY_PARMS *src, TPM_KEY_PARMS *dest)
{
  if (key_algorithmID_init_parms(src->algorithmID, dest))
    return -1;
  dest->encScheme = src->encScheme;
  dest->sigScheme = src->sigScheme;
  if (dest->parmFns->parms_init_algparms(src, dest->bparms)) {
    key_parms_free(dest);
    return -1;
  }
  return 0;
}


/**********************************************************************
 * .._parms_native_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the input parms and native-format keypair
 */
int key_parms_native_init_storepubkey(TPM_KEY_PARMS *parms, BYTE *native, 
				      TPM_STORE_PUBKEY *out)
{
  return parms->parmFns->parms_native_init_storepubkey(parms, native, out);
}


/**********************************************************************
 * .._keydata_init_storepubkey initializes the TPM_STORE_PUBKEY
 * structure from the provided TPM_KEY_DATA object
 */
int key_keydata_init_storepubkey(TPM_KEY_DATA *key, TPM_STORE_PUBKEY *out)
{
  TPM_KEY_PARMS *parms = &key->algorithmParms;
  return parms->parmFns->parms_native_init_storepubkey(parms, key->bnativekey, 
						       out);
}


/**********************************************************************
 * .._pubkeydata_init_pubkey initializes the TPM_PUBKEY (external std format
 * public key) from the provided TPM_PUBKEY_DATA (internal use format public
 * key)
 */
int key_pubkeydata_init_pubkey(TPM_PUBKEY_DATA *src, TPM_PUBKEY *dest)
{
  if (key_parms_init_parms(&src->algorithmParms, &dest->algorithmParms))
    return -1;

  if (src->algorithmParms.parmFns->parms_nativepk_init_storepubkey(
			   &src->algorithmParms, src->bnativepubkey, 
			   &dest->pubKey)) {
    free_TPM_KEY_PARMS(dest->algorithmParms);
    return -1;
  }
  return 0;
}


/**********************************************************************
 * .._keydata_init_pubkey initializes the TPM_PUBKEY (external std format
 * public key) from the provided TPM_KEY_DATA (internal use format keypair
 */
int key_keydata_init_pubkey(TPM_KEY_DATA *key, TPM_PUBKEY *out)
{
  TPM_KEY_PARMS *parms = &key->algorithmParms;

  if (key_parms_init_parms(parms, &out->algorithmParms))
    return -1;
  if (parms->parmFns->parms_native_init_storepubkey(parms, key->bnativekey, 
						    &out->pubKey)) {
    free_TPM_KEY_PARMS(out->algorithmParms);
    return -1;
  }
  return 0;
}


/**********************************************************************
 * .._pubkey_init_pubkeydata initializes the TPM_PUBKEY_DATA (internal
 * use format public key) from the provided TPM_PUBKEY (external std
 * format public key)
 */
int key_pubkey_init_pubkeydata(TPM_PUBKEY *src, TPM_PUBKEY_DATA *dest)
{
  dest->valid = TRUE;
  if (key_parms_init_parms(&src->algorithmParms, &dest->algorithmParms))
    return -1;

  if (src->algorithmParms.parmFns->pubkey_init_nativepk(src, 
							dest->bnativepubkey)) {
    free_TPM_KEY_PARMS(dest->algorithmParms);
    return -1;
  }
  return 0;
}


/**********************************************************************
 * .._keydata_init_pubkeydata initializes the TPM_PUBKEY_DATA (internal
 * use format public key) from the provided TPM_KEY_DATA (internal
 * format keypair)
 */
int key_keydata_init_pubkeydata(TPM_KEY_DATA *key, TPM_PUBKEY_DATA *out)
{
  out->valid = 1;
  if (key_parms_init_parms(&key->algorithmParms, &out->algorithmParms))
    return -1;
  if (key->algorithmParms.parmFns->keydata_init_nativepk(key, 
							 out->bnativepubkey)) {
    free_TPM_KEY_PARMS(out->algorithmParms);
    return -1;
  }
  return 0;
}


/**********************************************************************
 * .._parms_native_init_storeprivkey initializes a TPM_STORE_PRIVKEY
 * (external format public key storage format) from the provided native
 * format keypair
 */
int key_parms_native_init_storeprivkey(TPM_KEY_PARMS *parms, BYTE *nativeKey,
				       TPM_STORE_PRIVKEY *privKey)
{
  return parms->parmFns->parms_native_init_storeprivkey(parms, nativeKey,
							privKey);
}


/**********************************************************************
 * .._keydata_init_keydata initializes a TPM_KEY_DATA (internal use key
 * format) by copying an existing TPM_KEY_DATA object
 */
int key_keydata_init_keydata(TPM_KEY_DATA *in, TPM_KEY_DATA *out)
{
  TPM_KEY_PARMS *parms = &in->algorithmParms;
  memcpy(out, in, sizeof(TPM_KEY_DATA));
  if (key_parms_init_parms(parms, &out->algorithmParms))
    return -1;
  if (in->algorithmParms.parmFns->parms_native_init_native(parms, 
							   in->bnativekey,
							   out->bnativekey)) {
    free_TPM_KEY_PARMS(out->algorithmParms);
    return -1;
  }
  return 0;
}


/******** Creating basic native representations *********/


/**********************************************************************
 * .._parms_storekeys_init_native initializes two external format
 * storage structures, one with public key and one with the private
 * key, from a native format keypair
 */
int key_parms_storekeys_init_native(TPM_KEY_PARMS *parms, 
				      TPM_STORE_PUBKEY *pubKey, 
				      TPM_STORE_PRIVKEY *privKey,
				      BYTE *native)
{
  return parms->parmFns->parms_storekeys_init_native(parms, pubKey, 
						     privKey, native);
}


/******** Basic crypto operations  *********/


/**********************************************************************
 * .._parms_generate_native generates a new random native-format keypair
 * according to parameters passed in using the parms structure
 */
int key_parms_generate_native(TPM_KEY_PARMS *parms, BYTE *nativeout)
{
  return parms->parmFns->parms_generate_native(parms, nativeout);
}


/**********************************************************************
 * .._pubkeydata_encrypt does an encryption in the obvious way (all buffers
 * must be pre-allocated)
 */
int key_pubkeydata_encrypt(TPM_PUBKEY_DATA *key, BYTE *in, UINT32 in_size,
			   BYTE *enc, UINT32 *enc_size)
{
  return key->algorithmParms.parmFns->pubkeydata_encrypt(key, in, in_size,
							 enc, enc_size);
}


/**********************************************************************
 * .._keydata_decrypt does a decryption in the obvious way (all buffers
 * must be pre-allocated)
 */
int key_keydata_decrypt(TPM_KEY_DATA *key, BYTE *enc, UINT32 enc_size,
			BYTE *out, UINT32 *out_size)
{
  return key->algorithmParms.parmFns->keydata_decrypt(key, enc, enc_size, 
						      out, out_size);
}


/**********************************************************************
 * .._parms_native_sign does a signature in the obvious way (all
 * buffers must be pre-allocated) - usually a hash is passed in, but
 * if instead raw data is passed in that should be hashed first you
 * can set the "hashFirst" flag
 */
TPM_RESULT key_parms_native_sign(TPM_KEY_PARMS *parms, BYTE *nativeKey, 
			  const uint8_t *data, size_t data_len, uint8_t *sig)
{
  return parms->parmFns->parms_native_sign(parms,nativeKey,
					   data,data_len,TRUE,sig);
}


/**********************************************************************
 * .._keydata_sign does a signature in the obvious way - usually a
 * hash is passed in, but if instead raw data is passed in that should
 * be hashed first you can set the "hashFirst" flag
 *
 * Note that unlike the sign function above, the signature buffer
 * should not be pre-allocated - it is allocated here (and the caller
 * is responsible for freeing up the signature space after it is no
 * longer needed)
 */
TPM_RESULT key_keydata_sign(TPM_KEY_DATA *key, BYTE *inData, UINT32 inDataSize,
			    BOOL hashFirst, BYTE **sig, UINT32 *sigSize)
{
  int res;
  *sigSize = key_parms_get_sigsize(&key->algorithmParms);
  *sig = tpm_malloc(*sigSize);
  if (*sig == NULL) return TPM_NOSPACE;
  res = key->algorithmParms.parmFns->parms_native_sign(&key->algorithmParms,
		         key->bnativekey, inData, inDataSize, hashFirst, *sig);
  if (res) {
    tpm_free(*sig);
    *sig = NULL;
    return res;
  }
  return res;
}


/**********************************************************************
 * .._parms_nativepk_verify verifies a signature using the supplied public
 * key - usually a hash is passed in, but if instead raw data is passed in
 * that should be hashed first you can set the "hashFirst"  flag
 */
TPM_RESULT key_parms_nativepk_verify(TPM_KEY_PARMS *parms, BYTE *nativepk, 
                    BYTE *data, UINT32 dataSize, BOOL hashFirst, uint8_t *sig)
{
  return parms->parmFns->parms_nativepk_verify(parms, nativepk, data,
					       dataSize, hashFirst, sig);
}


/******** Marshal/Unmarshal/Size functions for key types *********/

/**********************************************************************
 * .._parms_native_marshal_size returns the size (in bytes) that is required
 * when a native-format keypair is marshalled.
 */
UINT32 key_parms_native_marshal_size(TPM_KEY_PARMS *parms, BYTE *native)
{
  return parms->parmFns->parms_native_marshal_size(parms, native);
}


/**********************************************************************
 * .._parms_marshal_native marshals the provided native-format keypair into
 * a generic binary blob
 */
int key_parms_marshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, UINT32 *length, 
			     BYTE *native)
{
  return parms->parmFns->parms_marshal_native(parms, ptr, length, native);
}


/**********************************************************************
 * .._parms_unmarshal_native unmarshals the provided blob into
 * a native-format keypair
 */
int key_parms_unmarshal_native(TPM_KEY_PARMS *parms, BYTE **ptr, UINT32 *length,
			       BYTE *native)
{
  return parms->parmFns->parms_unmarshal_native(parms, ptr, length, native);
}


/**********************************************************************
 * .._parms_nativepk_marshal_size returns the size (in bytes) that is required
 * when a native-format public key is marshalled.
 */
UINT32 key_parms_nativepk_marshal_size(TPM_KEY_PARMS *parms, BYTE *native)
{
  return parms->parmFns->parms_nativepk_marshal_size(parms, native);
}


/**********************************************************************
 * .._parms_marshal_nativepk marshals the provided native-format public key
 * into a generic binary blob
 */
int key_parms_marshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, UINT32 *length, 
			     BYTE *native)
{
  return parms->parmFns->parms_marshal_nativepk(parms, ptr, length, native);
}


/**********************************************************************
 * .._parms_unmarshal_nativepk unmarshals the provided blob into
 * a native-format public key
 */
int key_parms_unmarshal_nativepk(TPM_KEY_PARMS *parms, BYTE **ptr, 
				 UINT32 *length, BYTE *native)
{
  return parms->parmFns->parms_unmarshal_nativepk(parms, ptr, length, native);
}


/**********************************************************************
 * .._parms_marshal_algkeyparms marshals the algorithm-specific portion of
 * the keyparms
 */
int key_parms_marshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				  UINT32 *length)
{
  return parms->parmFns->parms_marshal_algkeyparms(parms, ptr, length);
}


/**********************************************************************
 * .._parms_unmarshal_algkeyparms unmarshals into the algorithm-specific
 * portion of the keyparms
 */
int key_parms_unmarshal_algkeyparms(TPM_KEY_PARMS *parms, BYTE **ptr,
				    UINT32 *length)
{
  return parms->parmFns->parms_unmarshal_algkeyparms(parms, ptr, length);
}


/**********  Free functions for key structures *************/


/**********************************************************************
 * .._parms_free frees up any dynamically-allocated space used in the
 * algorithm-specific keyparms
 */
void key_parms_free(TPM_KEY_PARMS *parms)
{
  if (parms->parmSize > 0) {
    parms->parmFns->parms_free(parms);
    parms->parmSize = 0;
  }
}


/**********************************************************************
 * .._parms_native_free frees up any dynamically-allocated space used in the
 * native-format keypair
 */
void key_parms_native_free(TPM_KEY_PARMS *parms, BYTE *pnativeKey)
{
  parms->parmFns->parms_native_free(parms, pnativeKey);
}


/**********************************************************************
 * .._parms_nativepk_free frees up any dynamically-allocated space used
 * in the native-format public key
 */
void key_parms_nativepk_free(TPM_KEY_PARMS *parms, BYTE *pnativepk)
{
  parms->parmFns->parms_nativepk_free(parms, pnativepk);
}


/**********************************************************************
 * .._selftest_all_keytypes runs consistency checks on the routines
 * for all loaded keytypes
 */
int key_selftest_all_keytypes(void)
{
  int res = 0;
  int i;
  for (i=1; i<key_numKeyTypes; i++) {  // Note: Don't test null keytype
    if (key_typeInfo[i]->selftest()) {
      debug("selftest failed for keytype %d", key_typeInfo[i]->algorithmID);
      res = -1;
    }
  }
  return res;
}


