/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

/*
 * Original plugin/extension code by Alex Kilgore and Stephen Tate.
 */

#ifndef _PLUGINS_H
#define _PLUGINS_H

#include "tpm_structures.h"
#include "tpm_marshalling.h"

/*
 * Basic function to initialize plugin system and load plugins
 */

void plugins_init(void);
void plugins_add(const char *plist);


/*
 * Structure to define a new command - must specify the ordinal, a name
 * to print in debugging output, the function to call for execution,
 * offsets to input and output parameters that are part of authentication
 * (skips over things like keyhandles), and flags (defined below).
 */
typedef struct commands_info_s {
  TPM_COMMAND_CODE ordinal;
  const char *debugname;
  TPM_RESULT (*execute_cmd)(TPM_REQUEST *req, TPM_RESPONSE *rsp);
  UINT32 in_param_offset;
  UINT32 out_param_offset;
  UINT32 flags;
} commands_info;

/*
 * useflags values determine whether the command is available when the TPM
 * is either deactivated or disabled
 */
#define COMMANDS_OK_DEACTIVATED 0x01
#define COMMANDS_OK_DISABLED 0x02

/*
 * Structure to define a set of commands to add - basically just adds the
 * number of commands to an array of commands_info structures.  Note that
 * the array must be in increasing order of ordinal value so that binary
 * search can be used.
 */
typedef struct command_set_s {
  UINT32 numCommands;
  commands_info *commands;
} command_set;

/*
 * Table and function to manage command sets
 */

extern command_set commands_sets[];
extern int commands_numSets;
void commands_add_set(UINT32 numCommands, commands_info *cmds);

void tpm_compute_in_param_digest(TPM_REQUEST *req);

/*
 * Functions and defs for handling persistent data in plugins.
 */


#define TPM_EXTENSION_TAG 0x1234

/*
 * Functions that must be defined for plugin-used persistent data are
 * to: initialize the native-format (internal) representation of
 * persistent data with default values, convert data (external) format
 * to native (internal), and convert native (internal) to data
 * (external).
 */

typedef struct tdTPM_EXTENSION_FUNCTIONS {
  BYTE* (*init_native)(void);
  BYTE* (*data_init_native)(BYTE *data, UINT32 size);
  int (*native_init_data)(BYTE *native, BYTE **data, UINT32 *size);
} TPM_EXTENSION_FUNCTIONS;


/*
 * TPM_EXTENSION_DATA maintains data for plugin-used persistent data.
 * If native format exists, should use that for operations, otherwise
 * convert data to native first.
 */

typedef struct tdTPM_EXTENSION_DATA {
  UINT16 ext_id;
  TPM_EXTENSION_FUNCTIONS *fns;
  UINT32 size;
  BYTE *data;
  int native_valid;
  BYTE *native;
  struct tdTPM_EXTENSION_DATA *next;
} TPM_EXTENSION_DATA;

extern int plugin_tpm_store_permanent_data(void);
extern int plugin_tpm_restore_permanent_data(void);
extern BYTE *plugin_get_permdata(UINT16 ext_id, TPM_EXTENSION_FUNCTIONS *functions);

extern TPM_EXTENSION_DATA *plugin_permanent_data;

#endif
