/* Software-based Trusted Platform Module (TPM) Emulator
 * Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "tpm_emulator.h"
#include "tpm_commands.h"
#include "tpm_marshalling.h"
#include "tpm_data.h"
#include "crypto/sha1.h"

/*
 * Credential Handling ([TPM_Part3], Section 14)
 * There are two create EK commands. The first matches the 1.1 functionality. 
 * The second provides the mechanism to enable revokeEK and provides 
 * FIPS 140-2 compatibility. 
 */

TPM_RESULT tpm_get_pubek(TPM_PUBKEY *pubEndorsementKey)
{
  if (tpmData.permanent.data.endorsementKey.payload != TPM_PT_ASYM)
      return TPM_NO_ENDORSEMENT;
  if (key_keydata_init_pubkey(&tpmData.permanent.data.endorsementKey,
			      pubEndorsementKey) < 0)
      return TPM_FAIL;
  return TPM_SUCCESS;
}

TPM_RESULT TPM_CreateEndorsementKeyPair(TPM_NONCE *antiReplay,
                                        TPM_KEY_PARMS *keyInfo, 
                                        TPM_PUBKEY *pubEndorsementKey, 
                                        TPM_DIGEST *checksum)
{
  info("TPM_CreateEndorsementKeyPair()");
  return TPM_DISABLED_CMD;
}

TPM_RESULT TPM_CreateRevocableEK(TPM_NONCE *antiReplay, TPM_KEY_PARMS *keyInfo,
                                 BOOL generateReset, TPM_NONCE *inputEKreset,  
                                 TPM_PUBKEY *pubEndorsementKey, 
                                 TPM_DIGEST *checksum, 
                                 TPM_NONCE *outputEKreset)
{
  TPM_KEY_DATA *ek = &tpmData.permanent.data.endorsementKey;
  TPM_RESULT res;
  info("TPM_CreateRevocableEK()");
  /* verify key parameters */
  if (ek->payload == TPM_PT_ASYM)
      return TPM_DISABLED_CMD;
  if (!key_parms_is_supported(keyInfo)
      || !key_parms_is_strong_size(keyInfo)
      || !key_parms_is_good_es(keyInfo))
      return TPM_BAD_KEY_PROPERTY;
  ek->payload = TPM_PT_ASYM;
  if (key_parms_init_parms(keyInfo, &ek->algorithmParms)) {
      ek->payload = TPM_PT_NONE;
      return TPM_FAIL;
  }
  ek->keyUsage = TPM_KEY_STORAGE;
  ek->keyFlags = TPM_KEY_FLAG_PCR_IGNORE;
  /* create endorsement key */
  if (key_parms_generate_native(&ek->algorithmParms, ek->bnativekey)) {
      free_TPM_KEY_PARMS(ek->algorithmParms);
      ek->payload = TPM_PT_NONE;
      return TPM_FAIL;
  }
  /* return PUBEK */
  res = tpm_get_pubek(pubEndorsementKey);
  if (res != TPM_SUCCESS) {
      free_TPM_KEY_DATA((*ek));
      ek->payload = TPM_PT_NONE;
      return res;
  }
  /* compute checksum */
  if (tpm_compute_pubkey_checksum(antiReplay, pubEndorsementKey, checksum)) {
      free_TPM_PUBKEY((*pubEndorsementKey));
      free_TPM_KEY_DATA((*ek));
      ek->payload = TPM_PT_NONE;
      return TPM_FAIL;
  }
  tpmData.permanent.flags.enableRevokeEK = TRUE;
  tpmData.permanent.flags.CEKPUsed = TRUE;
  if (generateReset) {
    tpm_get_random_bytes(tpmData.permanent.data.ekReset.nonce, 
      sizeof(tpmData.permanent.data.ekReset.nonce));
  } else {
    memcpy(&tpmData.permanent.data.ekReset, inputEKreset, sizeof(TPM_NONCE));
  }
  memcpy(outputEKreset, &tpmData.permanent.data.ekReset, sizeof(TPM_NONCE));
  /* Create TPM_PERMANENT_DATA->TPM_DAA_TPM_SEED from the TPM RNG */
  tpm_get_random_bytes(tpmData.permanent.data.tpmDAASeed.nonce, 
    sizeof(tpmData.permanent.data.tpmDAASeed.nonce));
  return TPM_SUCCESS;
}

TPM_RESULT TPM_RevokeTrust(TPM_NONCE *ekReset)
{
  info("TPM_RevokeTrust()");
  if (!tpmData.permanent.flags.enableRevokeEK) return TPM_FAIL;
  if (!tpm_get_physical_presence()) return TPM_BAD_PRESENCE;
  if (memcmp(ekReset, &tpmData.permanent.data.ekReset, 
             sizeof(TPM_NONCE))) return TPM_AUTHFAIL;
  tpm_owner_clear();
  free_TPM_KEY_DATA(tpmData.permanent.data.endorsementKey);
  tpmData.permanent.data.endorsementKey.payload = TPM_PT_NONE;
  /* Invalidate TPM_PERMANENT_DATA->tpmDAASeed */
  memset(tpmData.permanent.data.tpmDAASeed.nonce, 0, 
    sizeof(tpmData.permanent.data.tpmDAASeed.nonce));
  return TPM_SUCCESS;
}

TPM_RESULT TPM_ReadPubek(TPM_NONCE *antiReplay, TPM_PUBKEY *pubEndorsementKey, 
                         TPM_DIGEST *checksum)
{
  TPM_RESULT res;
  info("TPM_ReadPubek()");
  if (!tpmData.permanent.flags.readPubek) return TPM_DISABLED_CMD;
  /* get PUBEK */
  res = tpm_get_pubek(pubEndorsementKey);
  if (res != TPM_SUCCESS) return res; 
  /* compute checksum */
  if (tpm_compute_pubkey_checksum(antiReplay, pubEndorsementKey, checksum)) {
    tpm_free(pubEndorsementKey->pubKey.key);
    return TPM_FAIL;
  }
  return TPM_SUCCESS;
}

TPM_RESULT TPM_DisablePubekRead(TPM_AUTH *auth1)
{
  TPM_RESULT res;
  info("TPM_DisablePubekRead()");
  /* verify authorization */
  res = tpm_verify_auth(auth1, tpmData.permanent.data.ownerAuth, TPM_KH_OWNER);
  if (res != TPM_SUCCESS) return res;
  tpmData.permanent.flags.readPubek = FALSE;
  return TPM_SUCCESS;
}

TPM_RESULT TPM_OwnerReadInternalPub(TPM_KEY_HANDLE keyHandle, TPM_AUTH *auth1,  
                                    TPM_PUBKEY *publicPortion)
{
  TPM_RESULT res;
  TPM_KEY_DATA *srk = &tpmData.permanent.data.srk;
  info("TPM_OwnerReadInternalPub()");
  /* verify authorization */
  res = tpm_verify_auth(auth1, tpmData.permanent.data.ownerAuth, TPM_KH_OWNER);
  if (res != TPM_SUCCESS) return res;
  if (keyHandle == TPM_KH_EK) {
    return tpm_get_pubek(publicPortion);
  } else if (keyHandle == TPM_KH_SRK) {
    if (key_keydata_init_pubkey(srk, publicPortion) < 0)
      return TPM_FAIL;
    return TPM_SUCCESS;
  } else {
    return TPM_BAD_PARAMETER;
  }
}
