/* Software-based Trusted Platform Module (TPM) Emulator
 * Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <time.h>
#include <sys/time.h>
#include <errno.h>

#include "tpm_emulator.h"
#include "tpm_timing.h"

/*
 * This file contains code to support "timing profiles" for various
 * TPM hardware, and the ability to delay command returns by an
 * appopriate amount to mimic that hardware.  Delays are added by core
 * crypto functions (signing, decrypting, etc.) as well as basic
 * command support infrastructure (input/output marshalling, etc.), so
 * that if experimental new commands are implemented then the
 * resulting timing estimate should be a reasonable approximation of
 * the time that would be required by the modelled hardware for the
 * new command.  More information about this can be found in:
 *
 *   V. Gunupudi and S. R. Tate.  "Timing-Accurate TPM Simulation for What-If
 *   Explorations in Trusted Computing," Proceedings of the International
 *   Symposium on Performance Evaluation of Computer and Telecommunications
 *   Systems (SPECTS), 2010.
 */

struct TPMTimingConfig TPMTimingOps[TPMTIMING_FENCE+1] = {
    { "overconst", 0.0, 0 },
    { "overinrate", 0.0, 0 },
    { "overoutrate", 0.0, 0 },
    { "decrypt", 0.0, 0 },
    { "encrypt", 0.0, 0 },
    { "sign", 0.0, 0 },
    { "hashrate", 0.0, 0 },
    { NULL, 0.0, 0 }
};

static int hist[TPMTIMING_FENCE+1];

int TPMTimingEnabled = 0;

void TPMTiming_enable(void)
{
  TPMTimingEnabled = 1;
}

#if 0
/*
 * Printing "histograms" of basic operation counts was particularly
 * useful in initial explorations of modeling timing for various TPMs,
 * to see what the component operations were.  It's less useful for
 * day-to-day operation, so the call to this function is commented out
 * below - the code is still here in case it is useful at some future
 * time.  Reenable it by removing the "#if" and "#endif" surrounding
 * the code.
 */
static void TPMTiming_printHist(const char *label)
{
  int i;
  debug("Histogram %s\n", label);
  for (i=0; i<=TPMTIMING_FENCE; i++)
    if (hist[i] > 0)
      debug("TPMTiming-count %s(%c): %d\n", TPMTimingOps[i].name, 
	     'a'+i, hist[i]);
}
#endif

static void mymillisleep(double delay)
{
  struct timespec ts, trem;
  unsigned long long int insdelay = delay*1000000;
  ts.tv_sec = (long)(insdelay / 1000000000LL);
  ts.tv_nsec = (long)(insdelay % 1000000000LL);

  while ((nanosleep(&ts, &trem)==-1) && (errno == EINTR)) {
    ts.tv_sec = trem.tv_sec;
    ts.tv_nsec = trem.tv_nsec;
  }
}


/* Note:  All times are kept internally in milliseconds */

static double TPMTiming_startTime = 0.0;
static double accumDelay = 0.0;

/*
 * Note that the "addDelay" functions below don't actually add a delay,
 * but rather just accumlate a total delay time which is performed once,
 * at the end of any command execution, by calling TPMTiming_execDelay().
 */

void TPMTiming_startDelay(void)
{
  int i;

  if (TPMTimingEnabled) {
    struct timeval tv;
    gettimeofday(&tv, NULL);
    TPMTiming_startTime = tv.tv_sec*1000 + tv.tv_usec/1000;
    accumDelay = 0.0;
    for (i=0; i<=TPMTIMING_FENCE; i++)
      hist[i] = 0;
  }
}


void TPMTiming_execDelay(void)
{
  if (TPMTimingEnabled) {
    debug("Computed command time: %g ms\n", accumDelay);
    struct timeval tv;
    gettimeofday(&tv, NULL);
    accumDelay -= (tv.tv_sec*1000 + tv.tv_usec/1000 - TPMTiming_startTime);
    if (accumDelay > 0.0) {
      mymillisleep(accumDelay);
    }
    /* Uncomment the following line to print a count of individual operations */
    /* TPMTiming_printHist("Final Exec");  */
  }
}


void TPMTiming_addDelay1(enum TPMTimingOffsets off)
{
  if (TPMTimingEnabled) {
    hist[off]++;
    accumDelay += TPMTimingOps[off].value;
  }
}


void TPMTiming_addDelay2(enum TPMTimingOffsets off, int count)
{
  if (TPMTimingEnabled) {
    hist[off] += count;
    accumDelay += count*TPMTimingOps[off].value;
  }
}


/*
 *  Code below here is for reading timing profiles and initializing profile
 *  variables.
 */

static int TPMTiming_setValue(char *name, double val)
{
  int i;

  for (i=0; TPMTimingOps[i].name!=NULL; i++) {
    if (!strcmp(TPMTimingOps[i].name, name)) {
      if (TPMTimingOps[i].isSet) {
	info("TPMTIMING profile error:  %s multiply defined\n", name);
	return 0;
      } else {
	TPMTimingOps[i].value = val;
	TPMTimingOps[i].isSet = 1;
	return 1;
      }
    }
  }

  info("TPMTIMING profile error:  Invalid configuration variable: %s\n", name);
  return 0;
}


static int internal_readConfig(FILE *fp)
{
  static const int BUFFLEN = 500;
  char linebuffer[BUFFLEN];
  char name[BUFFLEN];
  char xtra[BUFFLEN];
  double val;
  int i, okFlag;
  int linenum = 1;

  while (fgets(linebuffer, sizeof(linebuffer), fp) != NULL) {
    /* Remove spaces at beginning of line */
    char *cp = linebuffer;
    while ((*cp != '\0') && (isspace(*cp)))
      cp++;

    if ((*cp != '\0') && (cp != linebuffer)) {
      char *tcp = linebuffer;
      while ((*tcp++ = *cp++))
	;
    }

    int readlen = strlen(linebuffer);
    if (linebuffer[readlen-1] != '\n') {
      info("TPMTIMING profile error: Line %d: Line too long\n", linenum);
      return 0;
    } else {
      linebuffer[--readlen] = '\0';
      if ((linebuffer[0] != '#') && (readlen > 0)) {
	/* We now have a trimmed, non-comment line - process it */
	int num = sscanf(linebuffer, "%[^:]: %lf %s", name, &val, xtra);
	if (num == 2) {
	  if (!TPMTiming_setValue(name, val))
	    return 0;
	} else {
	  info("TPMTIMING profile error: Line %d:  Syntax error\n", linenum);
	  return 0;
	}
      }
    }
    linenum++;
  }

  /* We've read the full timing profile - make sure all values are set */
  okFlag = 1;
  for (i=0; TPMTimingOps[i].name!=NULL; i++)
    if (!TPMTimingOps[i].isSet) {
      if (okFlag)
	strncpy(linebuffer, "TPMTIMING profile error - undefined params: ",
		sizeof(linebuffer));
      else
	strncat(linebuffer, ", ", sizeof(linebuffer)-strlen(linebuffer)-1);
      strncat(linebuffer, TPMTimingOps[i].name, 
	      sizeof(linebuffer)-strlen(linebuffer)-1);
      okFlag = 0;
    }

  if (!okFlag) {
    info("%s", linebuffer);
    return 0;
  }

  return 1;
}


int TPMTiming_readConfig(const char *filename)
{
  FILE *fp;
  int ret;
  int found = 0;

  // First try exactly as typed
  if ((fp=fopen(filename, "r")) != NULL)
    found = 1;

  // Next, if not an absolute path name, try install directory
  if (!found && filename[0] != '/') {
    const char *glue = "/etc/tpmsim/";
    char fullname[strlen(filename)+strlen(glue)+strlen(INSTALL_PREFIX)+1];
    fullname[0] = '\0';
    sprintf(fullname, "%s%s%s", INSTALL_PREFIX, glue, filename);
    if ((fp=fopen(fullname, "r")) != NULL)
      found = 1;
  }

  if (!found) {
    info("TPMTIMING error: Can't read configuration file: %s\n", filename);
    return 0;
  }

  if ((ret=internal_readConfig(fp)))
    TPMTiming_enable();

  fclose(fp);

  return ret;
}
