/* Software-based Trusted Platform Module (TPM) Emulator
 * Emulator Copyright (C) 2004-2010 Mario Strasser <mast@gmx.net>
 * Timing/plugin code Copyright (C) 2010-2012 Stephen R. Tate <srtate@uncg.edu>
 *
 * This module is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published
 * by the Free Software Foundation; either version 2 of the License,
 * or (at your option) any later version.
 *
 * This module is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include "tpm/tpm_structures.h"
#include "tpm/plugins.h"


#define COMMANDS_MAXSETS 100

command_set commands_sets[COMMANDS_MAXSETS];
int commands_numSets = 0;


void plugins_add(const char *plist)
{
  char *mlist;
  char *pch;
  void (*plugin_init)(void);

  if ((mlist=malloc(strlen(plist)+1)) == NULL) {
    perror("plugins_add()");
    exit(1);
  }
  strcpy(mlist, plist);

  // we need to break opt_load_plugins from a comma separated list
  // so first get a token of our comma separated list
  pch = strtok(mlist,",");
  while(pch != NULL) {
    void *dlhandle = NULL;
    int success = 0;
    char modname[100];
    // add one new keytype in, module_names is
    // the array of loaded module names, so each pointer
    // gets allocated to the length of the string token

    fprintf(stderr, "Trying to load plugin %s...\n", pch);
    // First try current directory
    snprintf(modname, sizeof(modname), "./libtpmsim_plugin_%s.so", pch);

    if ((dlhandle=dlopen(modname, RTLD_NOW|RTLD_LOCAL)) != NULL) {
      success = 1;
    }

    // If current directory didn't work, try regular load path
    if (!success) {
      snprintf(modname, sizeof(modname), "libtpmsim_plugin_%s.so", pch);

      if ((dlhandle=dlopen(modname, RTLD_NOW|RTLD_LOCAL)) != NULL) {
	success = 1;
      }
    }

    // Finally, try install directory 
    if (!success) {
      snprintf(modname, sizeof(modname), "%s/lib/tpmsim/libtpmsim_plugin_%s.so", INSTALL_PREFIX, pch);

      if ((dlhandle=dlopen(modname, RTLD_NOW|RTLD_LOCAL)) != NULL) {
	success = 1;
      }
    }

    if (!success) {
      fprintf(stderr, "TPMD startup error: Couldn't find plugin %s\n", pch);
      exit(1);
    }

    if ((plugin_init=dlsym(dlhandle, "plugin_init")) == NULL) {
      perror("failed to find plugin_init function");
      exit(1);
    }

    (*plugin_init)();

    pch = strtok(NULL,",");

  }
  free(mlist);
}


void commands_add_set(UINT32 numCommands, commands_info *cmds)
{
  if (commands_numSets < COMMANDS_MAXSETS) {
    UINT32 i, good=1;
    for (i=0; i<numCommands-1; i++) {
      if (cmds[i].ordinal > cmds[i+1].ordinal)
	good = 0;
    }
    if (good) {
      commands_sets[commands_numSets].numCommands = numCommands;
      commands_sets[commands_numSets].commands = cmds;
      commands_numSets++;
      info("Successfully added %d commands", numCommands);
    } else {
      info("ERROR loading commands!  Ordinals in table not sorted.");
    }
  } else {
    info("ERROR loading commands!  Command sets table is full.");
  }
}

/*
 * Functions for initializing base configuration (base keytypes and
 * standard TPM commands)
 */

void tpm_init_commands_standard(void);

void plugins_init(void)
{
  plugins_add("base");
  tpm_init_commands_standard();
}

